/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * vim: set ts=8 sts=4 et sw=4 tw=99:
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef jit_ppc64le_SharedICHelpers_ppc64le_h
#define jit_ppc64le_SharedICHelpers_ppc64le_h

#include "jit/BaselineFrame.h"
#include "jit/BaselineIC.h"
#include "jit/MacroAssembler.h"
#include "jit/SharedICRegisters.h"

namespace js {
namespace jit {

// Distance from sp to the top Value inside an IC stub (no return address on
// the stack on Power ISA).
static const size_t ICStackValueOffset = 0;

struct BaselineStubFrame {
    uintptr_t savedFrame;
    uintptr_t savedStub;
    uintptr_t returnAddress;
    uintptr_t descriptor;
};

inline void
EmitRestoreTailCallReg(MacroAssembler& masm)
{
    // No-op; LR is always the return address.
}

inline void
EmitRepushTailCallReg(MacroAssembler& masm)
{
    // No-op the second; LR is always the return address.
}

inline void
EmitCallIC(MacroAssembler& masm, CodeOffset* callOffset)
{
  // The stub pointer must already be in ICStubReg.
  // Load stubcode pointer from the ICStub.
  // R2 won't be active when we call ICs, so we can use it as scratch.
  masm.loadPtr(Address(ICStubReg, ICStub::offsetOfStubCode()), R2.scratchReg());

  // Call the stubcode via a direct jump-and-link
  masm.call(R2.scratchReg());
  *callOffset = CodeOffset(masm.currentOffset());
}

inline void
EmitReturnFromIC(MacroAssembler& masm)
{
    masm.as_blr();
}

inline void
EmitBaselineLeaveStubFrame(MacroAssembler& masm)
{
  masm.loadPtr(
      Address(FramePointer, BaselineStubFrameLayout::ICStubOffsetFromFP),
      ICStubReg);
  masm.movePtr(FramePointer, StackPointer);
  masm.Pop(FramePointer);

  // Load the return address into our GPR "mirror."
  masm.Pop(ICTailCallReg);
  // Move to LR for branching.
  masm.xs_mtlr(ICTailCallReg);

  // Discard the frame descriptor.
  {
    SecondScratchRegisterScope scratch2(masm);
    masm.Pop(scratch2);
  }
}

template <typename AddrType>
inline void
EmitPreBarrier(MacroAssembler& masm, const AddrType& addr, MIRType type)
{
    // Calls made in the prebarrier may clobber LR, so save it first.
    masm.xs_mflr(ScratchRegister);
    masm.push(ScratchRegister);
    masm.guardedCallPreBarrier(addr, type);
    masm.pop(ScratchRegister);
    masm.xs_mtlr(ScratchRegister);
}

inline void
EmitStubGuardFailure(MacroAssembler& masm)
{
    // Load next stub into ICStubReg
    masm.loadPtr(Address(ICStubReg, ICCacheIRStub::offsetOfNext()), ICStubReg);

    // Return address is already loaded, just jump to the next stubcode.
    masm.jump(Address(ICStubReg, ICStub::offsetOfStubCode()));
}

} // namespace jit
} // namespace js

#endif /* jit_ppc64le_SharedICHelpers_ppc64le_h */
